/*
 * SSMSimulator.java
 * Copyright (C) 2007 Lars Frantzen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * See the LICENSE file for the full license
 *
 * Created on 3. Januar 2007, 15:18
 *
 */

package info.frantzen.testing.ssmsimulator;

import info.frantzen.testing.ssmsimulator.interpreter.MyNode;
import info.frantzen.testing.ssmsimulator.interpreter.javacc.DumontParser;
import info.frantzen.testing.ssmsimulator.interpreter.javacc.DumontParser.RuntimeViolation;
import info.frantzen.testing.ssmsimulator.ssm.InteractionVariable;
import info.frantzen.testing.ssmsimulator.ssm.Location;
import info.frantzen.testing.ssmsimulator.ssm.Port;
import info.frantzen.testing.ssmsimulator.types.ComplexType;
import info.frantzen.testing.ssmsimulator.types.ComplexTypeInstance;
import info.frantzen.testing.ssmsimulator.types.Enumeration;
import info.frantzen.testing.ssmsimulator.types.ST_Boolean;
import info.frantzen.testing.ssmsimulator.types.ST_PosIntInstance;
import info.frantzen.testing.ssmsimulator.types.ST_PseudoPosDouble;
import info.frantzen.testing.ssmsimulator.types.ST_String;
import info.frantzen.testing.ssmsimulator.types.SimpleType;
import info.frantzen.testing.ssmsimulator.types.Type;
import info.frantzen.testing.ssmsimulator.types.TypeInstance;
import java.io.IOException;
import java.net.Socket;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import info.frantzen.testing.ssmsimulator.ssm.InstantiatedMessage;
import info.frantzen.testing.ssmsimulator.ssm.InstantiatedLocation;
import info.frantzen.testing.ssmsimulator.ssm.Message;
import info.frantzen.testing.ssmsimulator.ssm.MessageKind;
import info.frantzen.testing.ssmsimulator.ssm.ServiceStateMachine;
import info.frantzen.testing.ssmsimulator.ssm.Switch;
import info.frantzen.testing.ssmsimulator.ssm.Valuation;
import info.frantzen.testing.ssmsimulator.ssm.Variable;
import java.util.List;
import java.util.logging.Logger;

/**
 * An on-the-fly simulator for Service State Machines (SSM).<br>
 * <br>
 * The main simulation approach is to maintain a set of {@link InstantiatedLocation instantiated locations}
 * in which the SSM (which is passed to the constructor) could currently be.
 * <br>
 * This version supports the GNU Prolog treeSolver interface to compute input messages and
 * decide quiescence. The server must be running when creating the simulator, its
 * socket is a parameter of the constructor.
 * @author Lars Frantzen
 * @version 0.2
 */
public class SSMSimulator {
    
//    private static Logger logger = Logger.getLogger("info.frantzen.testing.ssmsimulator");
    private ServiceStateMachine ssm;
    private HashSet<InstantiatedLocation> instantiatedLocations;
    private HashSet<InstantiatedLocation> lastInstantiatedLocations;
    private TreeSolverInterface tsi;
    private boolean globalNewInputs = false;
    private boolean displayViaSocket = false;
    private boolean rememberSuspensionTrace = false;
    private QSDEInterface qsdei;
    private List<InstantiatedMessage> suspensionTrace = new ArrayList<InstantiatedMessage>();
    private Logger logger;
    
    /**
     *
     * Creates a new simulator for the passed SSM and computes the initial set of
     * possible instantiated locations.
     * @param ssm the SSM to be simulated
     * @throws java.io.IOException When the treeSolver cannot be reached via the socket, this exception is
     * thrown.
     * @throws ssmsimulator.interpreter.javacc.DumontParser.RuntimeViolation Already when constructing the SSMSimulator a RuntimeViolation may be thrown since
     * the simulator calculates the set of initial instantiated locations of the passed SSM.
     * By so doing it may has to execute update mappings and evaluate switch restrictions,
     * which in turn may throw a RuntimeViolation.
     */
    public SSMSimulator(ServiceStateMachine ssm, Socket solverSocket, Logger logger) throws RuntimeViolation, IOException {
        this.ssm = ssm;
        this.logger = logger;
        
        // Get initial instantiated location. No need to clone here since the following call to doInternalSteps
        // will just reassign these initial TypeInstances to the location variables.
        InstantiatedLocation initIL = ssm.getInstantiatedLocation();
        
        // We need a dummy Switch for the initial call to doInternalSteps with an empty update mapping.
        Switch dummySwitch = new Switch(new Message("", MessageKind.UNOBSERVABLE, null, null, null), "true == true", "");
        DumontParser dummyParser = null;
        try { dummyParser = ssm.makeParser("" , 1, dummySwitch); } catch (Exception ex) { ex.printStackTrace(); }
        dummySwitch.attachUMParser(dummyParser);
        
        // Compute the initial set of instantiated locations (since there could be unobservable switches originating
        // from the initial location)
        instantiatedLocations = doInternalSteps(initIL, dummySwitch);
        lastInstantiatedLocations = instantiatedLocations;
        
        // Create the interface to the solver. This demands a working socket connection.
        tsi = new TreeSolverInterface(solverSocket, logger);
        
        logger.info("SSMSimululator ready.");
    }
    
    private void backupInstantiatedLocations() {
        // hope a shallow copy suffices here
        lastInstantiatedLocations = (HashSet<InstantiatedLocation>) instantiatedLocations.clone();
    }
    
    public void undo() {
        instantiatedLocations = lastInstantiatedLocations;
    }
    
    public void setDisplaySocket(Socket qsdeSocket, String name) throws IOException {
        qsdei = new QSDEInterface(qsdeSocket, logger);
        qsdei.initialiseSequenceDiagram(name);
        displayViaSocket = true;
    }
    
    public void setRemeberSuspensionTrace(boolean b) {
        rememberSuspensionTrace = b;
    }
    
    public boolean getRememberSuspensionTraces() {
        return rememberSuspensionTrace;
    }
    
    public void setGlobalNewInputs(boolean b) {
        globalNewInputs = b;
    }
    
    public boolean getGlobalNewInputs() {
        return globalNewInputs;
    }
    
    /*
     * TODO: This method does not terminate if the SSM is not convergent, meaning that there is a loop of unobservable
     * switches.
     */
    private HashSet<InstantiatedLocation> doInternalSteps(InstantiatedLocation il, Switch aSwitch) throws RuntimeViolation {
        HashSet<InstantiatedLocation> result = new HashSet<InstantiatedLocation>();
        boolean onlyTaus = true;
        boolean allTausInfeasible = true;
        
        // We have visited the location and the switch
        il.getLocation().visited();
        aSwitch.visited();
        
        ssm.setInstantiatedLocation(il);
        
        // Execute the update mapping of the current switch
        aSwitch.executeUpdateMapping();
        
        Valuation currentValuation = ssm.getInstantiatedLocation().getValuation();
        
        for (Iterator i = ssm.getCurrentLocation().getOutgoingSwitches().iterator(); i.hasNext();) {
            Switch next = (Switch) i.next();
            if (next.getMessage().getKind() == MessageKind.UNOBSERVABLE)
                if (next.evaluateSwitchRestriction()) {
                allTausInfeasible = false;
                InstantiatedLocation nextIL = new InstantiatedLocation(next.getTarget(), currentValuation.clone());
                result.addAll(doInternalSteps(nextIL, next));
                } else
                    ;
            else
                onlyTaus = false;
        }
        
//        if (!onlyTaus || allTausInfeasible || currentLocation.getOutgoingSwitches().size() == 0) {
        if (!onlyTaus || allTausInfeasible) {
            result.add(new InstantiatedLocation(il.getLocation(), currentValuation));
        }
        
        return result;
    }
    
    /**
     * An instantiated message is simulated and the new set of possible instantiated locations is computed.
     * @param im the instantiated message to be processed
     * @throws ssmsimulator.SSMSimulator.EmptyInstantiatedLocationsException The resulting set of instantiated locations (after simulating the passed
     * instantied message) is empty.
     * @throws ssmsimulator.interpreter.javacc.DumontParser.RuntimeViolation A runtime violation of the parser.
     */
    public void processInstantiatedMessage(InstantiatedMessage im) throws EmptyInstantiatedLocationsException, RuntimeViolation {
        if (rememberSuspensionTrace)
            suspensionTrace.add(im);
        if (displayViaSocket)
            qsdei.updateSequenceDiagram(im);
        
        HashSet newInstantiatedLocations = new HashSet();
        
        for (Iterator i = instantiatedLocations.iterator(); i.hasNext();) {
            InstantiatedLocation is = (InstantiatedLocation) i.next();
            newInstantiatedLocations.addAll(after(is, im));
        }
        
        backupInstantiatedLocations();
        instantiatedLocations = newInstantiatedLocations;
        if (instantiatedLocations.isEmpty())
            fail();
    }
    
    
    public void processInstantiatedMessageNoBackup(InstantiatedMessage im) throws EmptyInstantiatedLocationsException, RuntimeViolation {
        if (rememberSuspensionTrace)
            suspensionTrace.add(im);
        if (displayViaSocket)
            qsdei.updateSequenceDiagram(im);
        
        HashSet newInstantiatedLocations = new HashSet();
        
        for (Iterator i = instantiatedLocations.iterator(); i.hasNext();) {
            InstantiatedLocation is = (InstantiatedLocation) i.next();
            newInstantiatedLocations.addAll(after(is, im));
        }
        
        // backupInstantiatedLocations();
        instantiatedLocations = newInstantiatedLocations;
        if (instantiatedLocations.isEmpty())
            fail();
    }
    
    
    // Premise: no non-covered tau actions from current location
    // This is the after for one stimulus or reaction, as in the FATES 04 paper
    private HashSet<InstantiatedLocation> after(InstantiatedLocation il, InstantiatedMessage im) throws RuntimeViolation {
        HashSet<InstantiatedLocation> result = new HashSet<InstantiatedLocation>();
        Message m = im.getMessage();
        Valuation val = im.getValuation();
        
        ssm.setInstantiatedLocation(il);
        
        // set new values for the interaction vars
        ssm.setInteractionVariables(val);
        
        for (Iterator i = ssm.getCurrentLocation().getOutgoingSwitches().iterator(); i.hasNext();) {
            Switch next = (Switch) i.next();
            if ((next.getMessage() == m) && next.evaluateSwitchRestriction()) {
                InstantiatedLocation nextIL = new InstantiatedLocation(next.getTarget(), il.getValuation());
                // System.out.println(next.getUpdateMapping());
                result.addAll(doInternalSteps(nextIL, next));
            }
        }
        return result;
    }
    
    /**
     * All non-quiescent instantiated locations are removed from the maintained
     * set of instantiated locations. Although there is a port needed as a parameter, the current simulator can only deal
     * with global quiecence!
     *
     * @throws ssmsimulator.SSMSimulator.EmptyInstantiatedLocationsException No current instantiated location was allowed to be quiescent.
     * @throws ssmsimulator.interpreter.javacc.DumontParser.RuntimeViolation A runtime violation of the parser.
     */
    public void processQuiescence(Port p) throws EmptyInstantiatedLocationsException, RuntimeViolation, IOException {
        if (rememberSuspensionTrace)
            suspensionTrace.add(p.getMyQuiescence());
        if (displayViaSocket)
            qsdei.updateSequenceDiagram(p.getMyQuiescence());
        
        HashSet<InstantiatedLocation> nonQuiescentLocations = new HashSet<InstantiatedLocation>();
        
        for (Iterator i = instantiatedLocations.iterator(); i.hasNext();) {
            InstantiatedLocation il = (InstantiatedLocation) i.next();
            if (!checkQuiescence(il))
                nonQuiescentLocations.add(il);
        }
        
        backupInstantiatedLocations();
        instantiatedLocations.removeAll(nonQuiescentLocations);
        if (instantiatedLocations.isEmpty())
            fail();
    }
    
    
    private boolean checkQuiescence(InstantiatedLocation il) throws RuntimeViolation, IOException {
        ssm.setInstantiatedLocation(il);
        
        for (Iterator i = ssm.getCurrentLocation().getOutgoingSwitches().iterator(); i.hasNext();) {
            Switch next = (Switch) i.next();
            
            if (! (next.getMessage().getKind() == MessageKind.INPUT))
                if (findSolution(next) != null)
                    return false;
        }
        return true;
    }
    
    /**
     * This method checks if the switch restriction of a given {@link Switch} has a solution.
     * If yes, an {@link InstantiatedMessage} for the switch with the computed solution
     * is returned. If not, null is returned. This method uses the Prolog treeSolver socket
     * connection.
     * @param s the switch for which a solution shall be found
     * @return an {@link InstantiatedMessage} if a solution exists, null otherwise
     * @throws info.frantzen.testing.ssmsimulator.interpreter.javacc.DumontParser.RuntimeViolation Not all types are supported, e.g. doubles cannot be solved. In this case
     * this exception is thrown.
     * @throws java.io.IOException If a socket problem occurs, this exception is thrown.
     */
    public InstantiatedMessage findSolution(Switch s) throws RuntimeViolation, IOException {
        Message message = s.getMessage();
        
        // The message is unobservable or has an empty type. Here we can simply evaluate the
        // switch restriction. If it is true, this switch can be taken, otherwise not
        if ((message.getKind() == MessageKind.UNOBSERVABLE) || message.getType().isEmpty()) {
            if (s.evaluateSwitchRestriction())
                return new InstantiatedMessage(message, new Valuation());
            else
                return null;
        }
        
        // Generate the Prolog query. This also shows which interaction variables are
        // actually present in the switch restriction
        String query = s.generatePrologQuery();
        
        // If there are no interaction vars in the switch restriction, and the
        // restriction evaluates to false, this switch cannot be taken, we return null
        if (MyNode.prologVarCounter == 0) {
            if (!s.evaluateSwitchRestriction())
                return null;
            else
                // If there are no interaction variable in the switch restriction, and the
                // restriction evaluates to true, this switch can be taken with random values
                // for the interaction variables
                return generateRandomMessage(message);
        } else {
            // there is at least one interaction var in the restriction
            System.out.println("Trying to solve: " + s.getSwitchRestriction());
            System.out.println("With query: " + query);
            // Solve the query
            Hashtable<String,Integer> prologSolution;
            // If we generate always new inputs, we ask the solver to generate a new solution. We do this only for
            // input switches since a solution for an output switch is only needed to check quiescence
            if ((globalNewInputs || s.getAlwaysNewInputs()) && s.getMessage().getKind() == MessageKind.INPUT) {
                prologSolution = tsi.askSolverAlwaysNew(MyNode.prologVarCounter, query, s.getFoundSolutions());
                if (prologSolution == null) { // We try to find a solution which has been computed before
                    if (s.getAlwaysNewDeleteSolutions())
                        s.clearFoundSolutions();
                    prologSolution = tsi.askSolver(MyNode.prologVarCounter, query);
                }
            } else
                prologSolution = tsi.askSolver(MyNode.prologVarCounter, query);
            // No solution found?
            if (prologSolution == null)
                return null;
            // Has a solution
            System.out.println("Solution found: " + prologSolution + " for " + MyNode.prologVartab);
            // If we generate always new inputs, we save the found solution in the switch
            if ((globalNewInputs || s.getAlwaysNewInputs()) && s.getMessage().getKind() == MessageKind.INPUT)
                s.addSolution(prologSolution);
            // For all interaction variable parts of the type: if it is in the restriction, take the
            // solver solution. Otherwise take a random solution.
            Valuation solutionValuation = new Valuation();
            for (Iterator it = message.getType().iterator(); it.hasNext();) {
                InteractionVariable var = (InteractionVariable) it.next();
                Type type = var.getType();
                TypeInstance instance = null;
                if (type instanceof ComplexType)
                    instance = createComplexInstance(type, prologSolution, var.getName());
                else if (type instanceof Enumeration)
                    instance = createEnumerationInstance(type, prologSolution, var.getName());
                else // (type instanceof SimpleType)
                    instance = createSimpleTypeInstance(type, prologSolution, var.getName());
                
                solutionValuation.addSingleValuation(var.getName(), instance);
            }
            
            InstantiatedMessage im = new InstantiatedMessage(message, solutionValuation);
            return im;
        }
    }
    
    private TypeInstance createComplexInstance(Type type, Hashtable prologSolution, String currentVarName) {
        if (type instanceof Enumeration)
            return createEnumerationInstance(type, prologSolution, currentVarName);
        if (type instanceof SimpleType)
            return createSimpleTypeInstance(type, prologSolution, currentVarName);
        
        // We have still a complex type
        Object[][] fields = ((ComplexType) type).getFields();
        Object[] resultingFieldValues = new Object[fields.length];
        
        for (int i=0; i<fields.length; i++) {
            String newCurrentName = currentVarName + '.' + (String) fields[i][0];
            Type newType = (Type) fields[i][1];
            resultingFieldValues[i] = createComplexInstance(newType, prologSolution, newCurrentName);
        }
        
        ComplexTypeInstance result = new ComplexTypeInstance(resultingFieldValues);
        return result;
    }
    
    private TypeInstance createEnumerationInstance(Type type, Hashtable prologSolution, String varname) {
        String pname = MyNode.prologVartab.get(varname);
        if (pname == null)   // The variable was not restricted
            return type.generateRandomInstance();
        else  // The variable was restricted and a solution has been found
            return ((Enumeration) type).instanceFromNumber((Integer) prologSolution.get(pname));
    }
    
    private TypeInstance createSimpleTypeInstance(Type type, Hashtable prologSolution, String varname) {
        SimpleType stype = (SimpleType) type;
        String restriction = stype.getRestriction();
        if (restriction == "Boolean") {
            String pname = MyNode.prologVartab.get(varname);
            if (pname == null)   // The variable was not restricted
                return stype.generateRandomInstance();
            else  // The variable was restricted and a solution has been found
                return ST_Boolean.instanceFromNumber((Integer) prologSolution.get(pname));
        } else if (restriction == "Integer") {
            String pname = MyNode.prologVartab.get(varname);
            if (pname == null)   // The variable was not restricted
                return stype.generateRandomInstance();
            else  // The variable was restricted and a solution has been found
                return new ST_PosIntInstance((Integer) prologSolution.get(pname));
        } else if (restriction == "Double") {
            String pname = MyNode.prologVartab.get(varname);
            if (pname == null)   // The variable was not restricted
                return stype.generateRandomInstance();
            else  // The variable was restricted and a solution has been found
                return ST_PseudoPosDouble.instanceFromNumber((Integer) prologSolution.get(pname));
        } else if (restriction == "String") {
            String pname = MyNode.prologVartab.get(varname);
            if (pname == null)   // The variable was not restricted
                return stype.generateRandomInstance();
            else  // The variable was restricted and a solution has been found
                return ST_String.instanceFromNumber((Integer) prologSolution.get(pname));
        }
        // should not come here
        return null;
    }
    
    
    /**
     * Returns all input switches departing from current instantiated locations.
     * @return all input switches departing from current instantiated locations
     */
    public HashSet<Switch> getCurrentInputSwitches() {
        return getCurrentSwitches(MessageKind.INPUT);
    }
    
    public HashSet<Switch> getCurrentOutputSwitches() {
        return getCurrentSwitches(MessageKind.OUTPUT);
    }
    
    private HashSet<Switch> getCurrentSwitches(MessageKind kind) {
        HashSet<Switch> result = new HashSet<Switch>();
        
        for (Iterator i = instantiatedLocations.iterator(); i.hasNext();) {
            InstantiatedLocation il = (InstantiatedLocation) i.next();
            
            // ssm.setInstantiatedLocation(il);
            
            //for (Iterator it = ssm.getCurrentLocation().getOutgoingSwitches().iterator(); it.hasNext();) {
            for (Iterator it = il.getLocation().getOutgoingSwitches().iterator(); it.hasNext();) {
                Switch next = (Switch) it.next();
                if ((next.getMessage().getKind() == kind)) {
                    result.add(next);
                }
            }
        }
        return result;
    }
    
    /**
     * Checks if a given switch together with an instantiated message can be taken
     * from one of the current instantiated locations.
     *
     * <B>Premise: the switch departs from at least one of the current instantiated locations.
     *    If not, false is returned. We don't check here if the (instantiated) message belongs to the Switch.
     *    This method may be changed/removed in the future, so it should better not be used.</B>
     * @param s the switch to be checked
     * @param inm the message for the switch
     * @return true, if the switch together with the message can be taken; false othewise
     * @throws ssmsimulator.interpreter.javacc.DumontParser.RuntimeViolation A runtime violation of the parser.
     */
    public boolean checkSwitchRestriction(Switch s, InstantiatedMessage inm) throws RuntimeViolation {
        for (Iterator it = instantiatedLocations.iterator() ; it.hasNext() ;) {
            InstantiatedLocation il = (InstantiatedLocation) it.next();
            if (il.getLocation().getOutgoingSwitches().contains(s)) {
                ssm.setInstantiatedLocation(il);
                ssm.setInteractionVariables(inm.getValuation());
                if (s.evaluateSwitchRestriction())
                    return true;
            }
        }
        return false;
    }
    
    /**
     * Generates a random instantiated message for a given Message.
     * @param m the message for which a random instantiation is needed
     * @return an instantiated message with random values
     */
    public InstantiatedMessage generateRandomMessage(Message m) {
        Valuation valu = new Valuation();
        
        // new random instances for all interaction variables
        for (Iterator it = m.getType().iterator(); it.hasNext() ;) {
            Variable ia = (Variable) it.next();
            valu.addSingleValuation(ia.getName(), ia.getType().generateRandomInstance());
        }
        return new InstantiatedMessage(m, valu);
    }
    
    /**
     * This method tries to find an instantiated message for a given switch which
     * respects the switch restriction. It does so by generating and checking random
     * instances. This will only succeed in very trivial cases. To do this in a
     * systematic way, use the {@link #findSolution} method.
     * @param s the switch for which an instantiated message is needed
     * @param maxtrials the maximum number of tried random instances
     * @return an instantiated messsage if one is randomly found, null otherwise
     * @throws ssmsimulator.interpreter.javacc.DumontParser.RuntimeViolation A runtime violation of the parser.
     */
    public InstantiatedMessage findRandomSolution(Switch s, int maxtrials) throws RuntimeViolation {
        for (int i=1; i <= maxtrials; i++) {
            InstantiatedMessage inm = generateRandomMessage(s.getMessage());
            // System.err.println(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>> Trying instance " + i + "\n" + inm.getValuation());
            if (checkSwitchRestriction(s, inm))
                return inm;
        }
        
        return null;
    }
    
    /**
     * A string describing the current set of instantiated locations.
     * @return a string describing the current set of instantiated locations.
     */
    public String getStatus() {
        StringBuffer sb = new StringBuffer();
        sb.append(instantiatedLocations.size() + " instantiated location(s):\n");
        for (Iterator i = instantiatedLocations.iterator(); i.hasNext();) {
            InstantiatedLocation il = (InstantiatedLocation) i.next();
            sb.append("Location " + il.getLocation() + ": " + il.getValuation());
            if (i.hasNext())
                sb.append('\n');
        }
        return sb.toString();
    }
    
    public double getLocationCoverage() {
        int locations = ssm.getLocations().size();
        int allvisits = 0;
        for (Iterator i = ssm.getLocations().iterator(); i.hasNext();) {
            Location il = (Location) i.next();
            if (il.getVistedCount() > 0)
                allvisits++;
        }
        return (allvisits / (double) locations) * 100;
    }
    
    public double getSwitchCoverage() {
        int switches = ssm.getSwitches().size();
        int allvisits = 0;
        for (Iterator i = ssm.getSwitches().iterator(); i.hasNext();) {
            Switch s = (Switch) i.next();
            if (s.getVistedCount() > 0)
                allvisits++;
        }
        return (allvisits / (double) switches) * 100;
    }
    
    public String getCoverageStatus() {
        StringBuffer sb = new StringBuffer();
        sb.append("Coverage Status\n---------------\n");
        int locations = ssm.getLocations().size();
        int visits = 0;
        int allvisits = 0;
        for (Iterator i = ssm.getLocations().iterator(); i.hasNext();) {
            Location il = (Location) i.next();
            visits = il.getVistedCount();
            sb.append("Location " + il + ": " + visits + " visit(s).\n");
            if (visits > 0)
                allvisits++;
        }
        double locationcov = (allvisits / (double) locations) * 100;
        sb.append(locationcov + " % location coverage.\n\n");
        
        int switches = ssm.getSwitches().size();
        visits = 0;
        allvisits = 0;
        for (Iterator i = ssm.getSwitches().iterator(); i.hasNext();) {
            Switch s = (Switch) i.next();
            visits = s.getVistedCount();
            sb.append("Switch " + s + ": " + visits + " visit(s).\n");
            if (visits > 0)
                allvisits++;
        }
        double switchcov = (allvisits / (double) switches) * 100;
        sb.append(switchcov + " % switch coverage.\n\n");
        
        return sb.toString();
    }
    
    /**
     * Getter for the current set of instantiated locations.
     * @return the current set of instantiated locations
     */
    public HashSet<InstantiatedLocation> getInstantiedLocations() {
        return instantiatedLocations;
    }
    
    private void fail() throws EmptyInstantiatedLocationsException {
        throw new EmptyInstantiatedLocationsException("No more instantiated locations!");
    }
    
    public List<InstantiatedMessage> getSuspensionTrace() {
        return suspensionTrace;
    }
    
    /**
     * This is an export to the Quick Sequence Diagram Editor. Only request-response and one-way is supported.
     * Every request must be answered by a response.
     */
    public String suspensionTraceAsQSDE() {
        if (rememberSuspensionTrace)
            return qsdei.suspensionTraceAsQSDE(suspensionTrace);
        else
            return "Suspension trace has not been remembered! Turn this feature on, first.";
    }
    
    public String asDotFormat() {
        return ssm.asDotDigraph();
    }
    
    /**
     * If a method leads to an empty set of instantiated locations, this exception is thrown.
     */
    public static class EmptyInstantiatedLocationsException extends Exception {
        EmptyInstantiatedLocationsException(String text) {
            super(text);
        }
    }
    
}
