/*
 * SimpleType.java
 * Copyright (C) 2007 Lars Frantzen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * See the LICENSE file for the full license
 *
 * Created on 29. Dezember 2006, 16:47
 *
 */

package info.frantzen.testing.ssmsimulator.types;

import javax.xml.namespace.QName;

/**
 * A simple type is a type wich directly maps to a primitive data type in Java,
 * referred to by what is called the "restriction" of the simple type.<br>
 * <br>
 * We always use the wrapper classes for the primitive types. You can restrict
 * Boolean, Integer, Double, and String.<br>
 * <br>
 * Furthermore is it possible to refer to just a subset of the domain of
 * the restricted primitive type, so you can for instance define
 * subcasses of this class describing positive doubles, odd integers,
 * strings of maximum size x, etc.<br>
 * <br>
 * This restriction is given by the predicate {@link #checkRestriction}.<br>
 * <br>
 * This class is abstract since no default implementations for the generation of instances can be given.
 * @author Lars Frantzen
 * @version 0.2
 */
public abstract class SimpleType extends Type {
    
    private String restriction;
    
    /**
     * This method checks if an object complies to the restriction of the simple type.
     * @param o the object to be checked
     * @return true, if the restriction holds, false otherwise
     */
    abstract public boolean checkRestriction(Object o);
    
    /**
     * Creates a new instance of SimpleType.
     * @param name the name of the simple type
     * @param restriction the primitive java type which the simple type restricts
     * (legal values are "Boolean", "Integer", "Double", and "String")
     */
    public SimpleType(QName name, String restriction) {
        super(name);
        this.restriction = restriction;
    }
    
    /**
     * Getter for the restriction (i.e. the primitive restricted java type as a sring).
     * @return the primitive restricted java type
     */
    public String getRestriction() {
        return restriction;
    }
    
    /**
     * Checks if the primitive restricted java type is one of "Boolean", "Integer", "Double", or "String".
     * @return true, if the consitenty holds, false otherwise
     */
    public boolean checkConsistency() {
        return ((restriction == "Boolean") || (restriction == "Integer") || (restriction == "Double") || (restriction == "String"));
    }
    
    public String getTypeIdentifier() {
        return this.toString();
    }
    
}
