/*
 * Type.java
 * Copyright (C) 2007 Lars Frantzen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * See the LICENSE file for the full license
 *
 * Created on 29. Dezember 2006, 16:47
 *
 */

package info.frantzen.testing.ssmsimulator.types;

import java.util.Random;
import javax.xml.namespace.QName;

/**
 * This is the abstract root class of the type system. The system supports three kinds of types:<br>
 * - {@link Enumeration enumeration} -  a list of string identifiers<br>
 * - {@link SimpleType simple types} - types which correspond to (a subset of) a java primitive type<br>
 * - {@link ComplexType complex types} - a type similar to an XML-Schema complex type, or what is called a
 * <I>record</I> in programming languages<br>
 * <br>
 * Every type has a corresponding {@link TypeInstance type instance} which represents values of
 * the given type.
 * @author Lars Frantzen
 * @version 0.2
 */
public abstract class Type {
    
    private QName name;
    /**
     * A random generator which can be used by subclasses to generate random values.
     */
    protected Random random = new Random();
    
    /**
     * Every concrete type has to provide a method to generate an initial type instance.
     * This is the same than initial values for Java variables (int is initially 0, an object is
     * initially null, etc.).
     * @return the initial type instance
     */
    public abstract TypeInstance generateInitialInstance();
    /**
     * Every concrete type has to provide a method to generate random type instances.
     * @return a random type instance
     */
    public abstract TypeInstance generateRandomInstance();
    /**
     * Every concrete type has to provide a method to generate the next boundary instance. Each
     * concrete type has to decide what is a boundary instance in its case.
     * @return the next boundary instance
     */
    public abstract TypeInstance generateNextBoundaryInstance();
    /**
     * Every concrete type has to provide a method returning the number of boundary instances. Each
     * concrete type has to decide what is a boundary instance in its case.
     * @return the number of boundary instances
     */
    public abstract int numberOfBoundaryInstances();
    /**
     * Every concrete type has to provide a method to do some consistency checks. Each
     * concrete type has to decide what is a consistency check in its case.
     * @return true, if the consistency check is ok, false otherwise
     */
    public abstract boolean checkConsistency();
    
    /**
     * Creates a new Type object.
     * @param name the qualified name of the type
     */
    public Type(QName name) {
        this.name = name;
    }
    
    /**
     * Getter for the name of the type.
     * @return the name of the type
     */
    public String getName() {
        return name.getLocalPart();
    }
    
    public abstract String getTypeIdentifier();
}
